@@ -1,11 +1,13 @@
 #!/usr/bin/env python
-
-from tests import unittest, as_s_expression
+import re
+import random
+import string
+import threading
+from tests import unittest, OrderedDict
 
 from jmespath import parser
+from jmespath import visitor
 from jmespath import ast
-from jmespath import lexer
-from jmespath import compat
 from jmespath import exceptions
 
 
@@ -13,13 +15,21 @@ class TestParser(unittest.TestCase):
     def setUp(self):
         self.parser = parser.Parser()
 
+    def assert_parsed_ast(self, expression, expected_ast):
+        parsed = self.parser.parse(expression)
+        self.assertEqual(parsed.parsed, expected_ast)
+
+    def test_parse_empty_string_raises_exception(self):
+        with self.assertRaises(exceptions.EmptyExpressionError):
+            self.parser.parse('')
+
     def test_field(self):
-        parsed = self.parser.parse('foo')
-        self.assertEqual(parsed.search({'foo': 'bar'}), 'bar')
+        self.assert_parsed_ast('foo', ast.field('foo'))
 
     def test_dot_syntax(self):
-        parsed = self.parser.parse('foo.bar')
-        self.assertEqual(parsed.search({'foo': {'bar': 'baz'}}), 'baz')
+        self.assert_parsed_ast('foo.bar',
+                               ast.subexpression([ast.field('foo'),
+                                                  ast.field('bar')]))
 
     def test_multiple_dots(self):
         parsed = self.parser.parse('foo.bar.baz')
@@ -33,10 +43,10 @@ class TestParser(unittest.TestCase):
             'one')
 
     def test_quoted_subexpression(self):
-        parsed = self.parser.parse('"foo"."bar"')
-        self.assertIsInstance(parsed.parsed, ast.SubExpression)
-        self.assertEqual(parsed.parsed.children[0].name, 'foo')
-        self.assertEqual(parsed.parsed.children[1].name, 'bar')
+        self.assert_parsed_ast('"foo"."bar"',
+                               ast.subexpression([
+                                   ast.field('foo'),
+                                   ast.field('bar')]))
 
     def test_wildcard(self):
         parsed = self.parser.parse('foo[*]')
@@ -64,19 +74,11 @@ class TestParser(unittest.TestCase):
         self.assertEqual(parsed.search({'foo': {'baz': 'baz'}}), None)
 
     def test_or_repr(self):
-        parsed = self.parser.parse('foo || bar')
-        self.assertEqual(repr(parsed), 'ORExpression(Field(foo), Field(bar))')
+        self.assert_parsed_ast('foo || bar', ast.or_expression(ast.field('foo'),
+                                                               ast.field('bar')))
 
     def test_unicode_literals_escaped(self):
-        parsed = self.parser.parse(r'`"\u2713"`')
-        if compat.PY2:
-            self.assertEqual(repr(parsed), r'Literal(\u2713)')
-        else:
-            self.assertEqual(repr(parsed), u'Literal(\u2713)')
-
-    def test_unicode_pretty_print(self):
-        parsed = self.parser.parse(r'`"\u2713"`')
-        self.assertEqual(parsed.pretty_print(), u'Literal(\u2713)')
+        self.assert_parsed_ast(r'`"\u2713"`', ast.literal(u'\u2713'))
 
     def test_multiselect(self):
         parsed = self.parser.parse('foo.{bar: bar,baz: baz}')
@@ -86,7 +88,6 @@ class TestParser(unittest.TestCase):
 
     def test_multiselect_subexpressions(self):
         parsed = self.parser.parse('foo.{"bar.baz": bar.baz, qux: qux}')
-        foo = parsed.search({'foo': {'bar': {'baz': 'CORRECT'}, 'qux': 'qux'}})
         self.assertEqual(
             parsed.search({'foo': {'bar': {'baz': 'CORRECT'}, 'qux': 'qux'}}),
             {'bar.baz': 'CORRECT', 'qux': 'qux'})
@@ -96,6 +97,15 @@ class TestParser(unittest.TestCase):
         result = parsed.search({'foo': {'bar': {'baz': 'CORRECT'}, 'qux': 'qux'}})
         self.assertEqual(result, {"bar": "CORRECT", "qux": "qux"})
 
+    def test_function_call_with_and_statement(self):
+        self.assert_parsed_ast(
+            'f(@ && @)',
+            {'children': [{'children': [{'children': [], 'type': 'current'},
+                                        {'children': [], 'type': 'current'}],
+                           'type': 'and_expression'}],
+             'type': 'function_expression',
+             'value': 'f'})
+
 
 class TestErrorMessages(unittest.TestCase):
 
@@ -119,12 +129,12 @@ class TestErrorMessages(unittest.TestCase):
 
     def test_bad_parse(self):
         with self.assertRaises(exceptions.ParseError):
-            parsed = self.parser.parse('foo]baz')
+            self.parser.parse('foo]baz')
 
     def test_bad_parse_error_message(self):
         error_message = (
-            'Unexpected token: ]: Parse error at column 3 '
-            'near token "]" (RBRACKET) for expression:\n'
+            'Unexpected token: ]: Parse error at column 3, '
+            'token "]" (RBRACKET), for expression:\n'
             '"foo]baz"\n'
             '    ^')
         self.assert_error_message('foo]baz', error_message)
@@ -136,27 +146,31 @@ class TestErrorMessages(unittest.TestCase):
             '                       ^')
         self.assert_error_message('foo.{bar: baz,bar: bar', error_message)
 
+    def test_incomplete_expression_with_missing_paren(self):
+        error_message = (
+            'Invalid jmespath expression: Incomplete expression:\n'
+            '"length(@,"\n'
+            '          ^')
+        self.assert_error_message('length(@,', error_message)
+
     def test_bad_lexer_values(self):
         error_message = (
             'Bad jmespath expression: '
-            'Starting quote is missing the ending quote:\n'
+            'Unclosed " delimiter:\n'
             'foo."bar\n'
             '    ^')
         self.assert_error_message('foo."bar', error_message,
                                   exception=exceptions.LexerError)
 
-    def test_bad_lexer_literal_value_with_json_object(self):
-        error_message = ('Bad jmespath expression: '
-                         'Bad token `{{}`:\n`{{}`\n^')
-        self.assert_error_message('`{{}`', error_message,
-                                  exception=exceptions.LexerError)
-
-
     def test_bad_unicode_string(self):
-        error_message = ('Bad jmespath expression: '
-                         'Invalid \\uXXXX escape:\n"\\uAZ12"\n^')
-        self.assert_error_message(r'"\uAZ12"', error_message,
-                                  exception=exceptions.LexerError)
+        # This error message is straight from the JSON parser
+        # and pypy has a slightly different error message,
+        # so we're not using assert_error_message.
+        error_message = re.compile(
+            r'Bad jmespath expression: '
+            r'Invalid \\uXXXX escape.*\\uAZ12', re.DOTALL)
+        with self.assertRaisesRegex(exceptions.LexerError, error_message):
+            self.parser.parse(r'"\uAZ12"')
 
 
 class TestParserWildcards(unittest.TestCase):
@@ -229,7 +243,7 @@ class TestParserWildcards(unittest.TestCase):
 
     def test_escape_sequence_at_end_of_string_not_allowed(self):
         with self.assertRaises(ValueError):
-            parsed = self.parser.parse('foobar\\')
+            self.parser.parse('foobar\\')
 
     def test_wildcard_with_multiselect(self):
         parsed = self.parser.parse('foo.*.{a: a, b: b}')
@@ -252,6 +266,7 @@ class TestParserWildcards(unittest.TestCase):
         self.assertIn('a', match[1])
         self.assertIn('b', match[1])
 
+
 class TestMergedLists(unittest.TestCase):
     def setUp(self):
         self.parser = parser.Parser()
@@ -294,10 +309,9 @@ class TestParserCaching(unittest.TestCase):
         for i in range(parser.Parser._MAX_SIZE + 1):
             compiled2.append(p.parse('foo%s' % i))
         self.assertEqual(len(compiled), len(compiled2))
-        s = as_s_expression
         self.assertEqual(
-            [s(expr.parsed) for expr in compiled],
-            [s(expr.parsed) for expr in compiled2])
+            [expr.parsed for expr in compiled],
+            [expr.parsed for expr in compiled2])
 
     def test_cache_purge(self):
         p = parser.Parser()
@@ -305,10 +319,34 @@ class TestParserCaching(unittest.TestCase):
         cached = p.parse('foo')
         p.purge()
         second = p.parse('foo')
-        self.assertEqual(as_s_expression(first.parsed),
-                         as_s_expression(second.parsed))
-        self.assertEqual(as_s_expression(first.parsed),
-                         as_s_expression(cached.parsed))
+        self.assertEqual(first.parsed,
+                         second.parsed)
+        self.assertEqual(first.parsed,
+                         cached.parsed)
+
+    def test_thread_safety_of_cache(self):
+        errors = []
+        expressions = [
+            ''.join(random.choice(string.ascii_letters) for _ in range(3))
+            for _ in range(2000)
+        ]
+        def worker():
+            p = parser.Parser()
+            for expression in expressions:
+                try:
+                    p.parse(expression)
+                except Exception as e:
+                    errors.append(e)
+
+        threads = []
+        for i in range(10):
+            threads.append(threading.Thread(target=worker))
+        for thread in threads:
+            thread.start()
+        for thread in threads:
+            thread.join()
+
+        self.assertEqual(errors, [])
 
 
 class TestParserAddsExpressionAttribute(unittest.TestCase):
@@ -318,5 +356,39 @@ class TestParserAddsExpressionAttribute(unittest.TestCase):
         self.assertEqual(parsed.expression, 'foo.bar')
 
 
+class TestParsedResultAddsOptions(unittest.TestCase):
+    def test_can_have_ordered_dict(self):
+        p = parser.Parser()
+        parsed = p.parse('{a: a, b: b, c: c}')
+        options = visitor.Options(dict_cls=OrderedDict)
+        result = parsed.search(
+            {"c": "c", "b": "b", "a": "a"}, options=options)
+        # The order should be 'a', 'b' because we're using an
+        # OrderedDict
+        self.assertEqual(list(result), ['a', 'b', 'c'])
+
+
+class TestRenderGraphvizFile(unittest.TestCase):
+    def test_dot_file_rendered(self):
+        p = parser.Parser()
+        result = p.parse('foo')
+        dot_contents = result._render_dot_file()
+        self.assertEqual(dot_contents,
+                         'digraph AST {\nfield1 [label="field(foo)"]\n}')
+
+    def test_dot_file_subexpr(self):
+        p = parser.Parser()
+        result = p.parse('foo.bar')
+        dot_contents = result._render_dot_file()
+        self.assertEqual(
+            dot_contents,
+            'digraph AST {\n'
+            'subexpression1 [label="subexpression()"]\n'
+            '  subexpression1 -> field2\n'
+            'field2 [label="field(foo)"]\n'
+            '  subexpression1 -> field3\n'
+            'field3 [label="field(bar)"]\n}')
+
+
 if __name__ == '__main__':
     unittest.main()
